/*
    Jun08,2000: sleep in mymalloc added, if temporarely not enough memory
    can give problems if parallelized!
    
 ToDo: 
       - replace mymalloc(...,NULL) calls
       - replace mymalloc, myfree by remalloc(0,0,...)

  v2.33 - bug fixed, mymap called with len>2^31 on 32bit systems
          also writing 2GB in single char steps by each MPI rank kills NFS
  v1.2 - fallback from malloc to mmap added (tmpfile if not enough mem) 

 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h> /* lseek() */
#include <unistd.h>    /* sleep() */
#include "defs.h" /* ULong */
#include "mem.h"

/* some external vars, not needed to include big h-files */
extern ULong usemem,maxmem;
extern int vvv;   /* verbose mode */
extern int mpi_i; /* mpi node */

minfo_t test0 = { NULL, 0, 0, "tmp/t_XXXXXX", NULL };

/* return 0 on success (old was: manipulating void **pp) */
void *myfree(void *pp, minfo_t *mi){ 
    if (!pp) return pp;
    if (mi) {
      usemem-=min(mi->l,usemem);
      if (vvv&2 && !mpi_i)
         printf(((usemem<9999999l)?
                "\n  free(%7.0f) -= %7.0f => %s"
               :"\n  free(%7.1e) -= %7.1e => %s"),
               (double)mi->l, (double)usemem, ((mi)?mi->name:""));
      if (pp!=mi->p) E0("mi.p != pp");
      if (!mi->l)    E0("mi.l == 0");
    }
    if (mi && mi->ff) {
      myunmap( pp,mi->ff,(long)mi->l,0);
      fclose(mi->ff); mi->ff=NULL;
    } else {
      free(pp);
    }
    if (mi) { mi->p=NULL; mi->l=0; }
    return NULL;
}

void *myremalloc(void *p0,unsigned long ll, minfo_t *mi){
    int itry; void *pp=NULL;

    if (ll==0l) { return myfree(p0,mi); }

    if (mi && !mi->p) mi->l=0;
    
    if ( (!mi) || (!mi->ff) )                  /* not mmapped */
    if ( usemem+((mi)?ll-mi->l:0) <= maxmem
      || maxmem==0l || (mi && (mi->mode&4)))
    for(itry=1;itry<5 && pp==NULL;itry++){    /* try for 120m=2h */
      pp = (void *)realloc(p0,(size_t)ll);
      /* if realloc failed? ret==0 => data untouched (not freed) */
      if (pp!=NULL || maxmem==0l) break;
      printf("\n alloc( %10p,%.0e) => %s failed, sleep & repeat",
             p0,(double)ll,((mi)?mi->name:""));
      sleep(itry*120); /* 120sec */
    }
    if (mi) {
#ifdef HAVE_MMAP
      if (!pp) { /* try mmap */
        if (!mi->ff) {
          mi->ff=fopen(mi->name,"r+b"); if(!mi->ff)
          mi->ff=fopen(mi->name,"w+b"); /* create */
          if (p0 && mi->l) 
           if (fwrite(p0,1,mi->l,mi->ff)!=mi->l) F1("write %s",mi->name);
          if (p0) free(p0);
          if (mi->ff) { pp=mymap(mi->name,mi->ff,(long)ll,0); }
          if (pp) { mi->l=ll; }
        } else { /* remap */
          itry=vvv; vvv=0; /* be quiet */
          myunmap( p0      ,mi->ff,(long)mi->l,0); mi->l=0;
          pp=mymap(mi->name,mi->ff,(long)ll,0);
          vvv=itry;
        }
      }
#endif
      if (pp && !mi->ff) usemem += ll - mi->l;
      if((vvv&2) && p0==NULL && !mpi_i) {
        /* %11.6g means print 6 digits number or (6+5) e-number */
        /* v2.32 dont print the pointer, better diff */
        printf(
           ((usemem>9999999l)?"\n alloc(%7.1e) += %7.1e => %s node= %2d"
                             :"\n alloc(%7.0f) += %7.0f => %s node= %2d"),
           (double)ll, (double)usemem, ((mi)?mi->name:""), mpi_i);
      }
      if (pp) { mi->p=pp; mi->l=ll; }
      if (!(mi->mode&1)) return pp;
    }
    if (!pp) { F2("realloc failed(%p,%ld), aborting",p0,(long)ll); }
    return pp;
}
    
void *mymalloc( unsigned long ll, minfo_t *mi){
    return myremalloc(NULL, ll, mi);
}

/* -------------------------------------------------------------------
    using files as memory to handle arrays bigger than mem+swap
    see libc.info (Memory-mapped I/O) for more information
    see utils/mmap1.c too
    how to use LFS (64bit)?
   ------------------------------------------------------------------- */
#define DEBUG 1

#ifdef HAVE_MMAP      /* else use malloc */
#include <unistd.h>   /* lseek (off_t = 64bit on LFS) */
#include <sys/mman.h> /* mmap */

static size_t page_size=0;
#endif

/*  map a memory to a file */
/*
 * fseek(long)  32bit or 64bit (v2.32 and less)
 * lseek(off_t) 32bit or 64bit (also 64bit for 32bit + LFS) v2.33
 *
 */
void *mymap(char *nam, FILE *f1, off_t len, off_t ofs){
  int fd=-1; off_t i, pos; void *buf; char *dummy;
  char    *s1[2]={"alloc+read","mmap"};
  char *s2=s1[0];
  if(len==0) return NULL;
  fd=fileno(f1);
#ifdef HAVE_MMAP
  /* systems have very different demands for offset in mmap function */
  /* here I try to find best for all systems */
#ifdef CFG_PAGESIZE /* this is not defined on every system (iMAC?) */
  page_size= (size_t) sysconf(_SC_PAGESIZE);
#else   /* if trouble: ./configure --mmap */
  page_size= 8192;  /* should work mostly, best (2^n) : n>=14 */
#endif
#endif
  for(;;) {
    pos=lseek(fd,(off_t)0,SEEK_END); /* fseek after file end failes */
    if (pos==-1) ERR(9999);
    if (pos>=ofs+len || mpi_i==mpi_n-1) break; /* last thread acts on EOF!? */
    sleep(10);  /* let wait task 1..n */
  }
  /* trouble if multiple filled up by mpi_i>0 */
  if (pos<ofs+len) { /* expand file by mpi_i=0 */
    dummy = (char*) malloc(8192);
    if (!dummy) ERR(9999);
    for (i=0; i<8192; i++) dummy[i]=0;
    for(i=0;i<(ofs+len-pos)/8192;i++) write(fd,dummy,8192); /* fill up */
    if (      (ofs+len-pos)%8192)     write(fd,dummy,(ofs+len-pos)%8192);
    fflush(f1); /* expand file */ 
    free(dummy);
  }
  /* the file must have already the size of the map and opened for r/w */
#ifdef HAVE_MMAP
  /* ofs must be a multiple of getpagesize() !!! */
  /* but most system dont care about that, on trouble do: configure --nomap */
  /* ToDo: use CFG_PAGESIZE defined by configure to fix that problem */
  if( page_size<=(size_t)len )
    buf=mmap(NULL,(size_t)len,PROT_READ | PROT_WRITE,MAP_SHARED,
             fd, (off_t)ofs);
    s2=s1[1]; /* really mmapped */
  else 
#endif
  {
    buf=malloc((size_t)len);
    if(buf==NULL) E1("malloc(len=%.0e)",(double)len);
    lseek(fd,ofs,SEEK_SET); pos=read(fd,buf,len); /* ssize_t, size_t */
    if( pos!=len ) E0("my_map: read failed");
    lseek(fd,  0,SEEK_END); /* lseek after file end failes */
  }
  /* v2.32 dont print the pointer to make diffs more usefull */
  if(vvv&2 && (mpi_i<2 || mpi_i>=mpi_n-2)){
    printf( ((ofs<9999999l && len<9999999l)?
         " mymap=%s(%7.0f) ofs=%9.0f %s fd=%2d mpi=%2d\n"
        :" mymap=%s(%7.1e) ofs=%9.3e %s fd=%2d mpi=%2d\n"),
          s2, 1.*len, 1.*ofs, nam, fd, mpi_i);
  }
  /* on error fall back to malloc+load (save+free) */
  if (buf==(void *)-1) {
     E3("mmap(fd=%d,len=%.2e,ofs=%.2e)",fd,1.*len,1.*ofs); perror(" "); }
  return buf;
}

int myunmap(void *buf, FILE *f1, off_t len, off_t ofs){
  int rc=0; off_t pos; int fd=fileno(f1);
  if (buf==NULL || buf==(void *)-1 || len==0) return 0;
  if (vvv&2 && (mpi_i<2 || mpi_i>=mpi_n-2)) {
    printf( ((ofs<9999999l && len<9999999l)?
       "\n myunmap(%7.0f) ofs=%7.0f fd=%2d"
      :"\n myunmap(%7.1e) ofs=%7.1e fd=%2d"),
        (double)len, (double)ofs, fileno(f1));
  }
  if (!f1)  E0("no open FILE");
#ifdef HAVE_MMAP
  if ( page_size<=(size_t)len ){
    rc=msync((caddr_t)buf,(size_t)len,MS_SYNC);
      if (rc) fprintf(stderr,"ERROR msync=%d\n",rc);
    rc=munmap((caddr_t)buf,(size_t)len);
      if (rc) fprintf(stderr,"ERROR munmap=%d\n",rc);
  } else
#endif
  {  /* save! ??? */
    if (mpi_n==1) {  /* v2.32 avoid file conflicts writing multiple l1 */
      if (vvv&2 && !mpi_i) printf(" write_to_file");
      lseek(fd,ofs,SEEK_SET); pos=write(fd,buf,len);
      if (pos!=len) {  /* more output, errors for MPI */
        printf("\nERROR my_unmap: fwrite failed for node=%d"
               " fd=%d ofs,len,written %.2e %.2e %.2e",
                mpi_i, fd, 1.*ofs, 1.*len, 1.*pos);
        perror(" PERROR=");
      }
    }
    if (buf) free(buf);
  }
  return rc;
}
